from importlib.metadata import version
from .utils import check_last_version, sendtosuperuser,compare_version
from nonebot.log import logger
import time
class Version():
    version:str#当前版本
    lastcheck: float = 0#上次检查时间
    ispushed:bool =True#是否已经推送
    latest:str="0.0.0"#最新版本
    package="nonebot-plugin-novelai"
    url="https://sena-nana.github.io/MutsukiDocs/update/novelai/"

    def __init__(self):
        #初始化当前版本
        try:
            self.version = version(self.package)
        except:
            self.version = "0.4.11"

    async def check_update(self):
        """检查更新，并推送"""
        #每日检查
        if time.time()-self.lastcheck > 80000:
            update = await check_last_version(self.package)
            #判断是否重复检查
            if await compare_version(self.latest,update):
                self.latest=update
                #判断是否是新版本
                if await compare_version(self.version,self.latest):
                    logger.info(self.push_txt())
                    self.ispushed=False
                else:
                    logger.info(f"novelai插件检查版本完成，当前版本{self.version}，最新版本{self.latest}")
            else:
                logger.info(f"novelai插件检查版本完成，当前版本{self.version}，最新版本{self.latest}")
            self.lastcheck = time.time()
        #如果没有推送，则启动推送流程
        if not self.ispushed:
            await sendtosuperuser(self.push_txt())
            self.ispushed=True

    def push_txt(self):
        #获取推送文本
        logger.debug(self.__dict__)
        return f"novelai插件检测到新版本{self.latest},当前版本{self.version},请使用pip install --upgrade {self.package}命令升级,更新日志：{self.url}"

version=Version()


from pydantic import BaseSettings, validator
from nonebot import get_driver
from nonebot.log import logger
from pydantic.fields import ModelField
from pathlib import Path
import aiofiles
import json
jsonpath = Path("data/novelai/config.json").resolve()


class Config(BaseSettings):
    novelai_token: str = ""  # 官网的token
    novelai_tag: str = ""  # 内置的tag
    novelai_uc: str = ""  # 内置的反tag
    novelai_cd: int = 60  # 默认的cd
    novelai_pure: bool = False  # 是否启用简洁返回模式（只返回图片，不返回tag等数据）
    novelai_limit: bool = True  # 是否开启限速
    novelai_save_pic: bool = True  # 是否保存图片至本地
    novelai_save_detail: bool = False  # 是否保存图片信息至本地
    novelai_api_domain: str = "https://api.novelai.net/"
    novelai_site_domain: str = "https://novelai.net/"
    novelai_mode: str = "novelai"
    novelai_paid: int = 0  # 0为禁用付费模式，1为点数制，2为不限制
    novelai_on: bool = True  # 是否全局开启
    novelai_h: bool = False  # 是否允许H
    novelai_oncemax: int = 3  # 每次能够生成的最大数量
    bing_key: str = None  # bing的翻译key
    deepl_key: str = None  # deepL的翻译key

    # 允许单群设置的设置
    def keys(cls):
        return ("novelai_cd", "novelai_tag", "novelai_on", "novelai_uc","novelai_pure")
    def __getitem__(cls,item):
        return getattr(cls,item)

    @validator("novelai_cd", "novelai_oncemax")
    def non_negative(cls, v: int, field: ModelField):
        if v < 1:
            return field.default
        return v

    @validator("novelai_paid")
    def paid(cls, v: int, field: ModelField):
        if v < 0:
            return field.default
        elif v > 2:
            return field.default
        return v

    class Config:
        extra = "ignore"

    def check_mode(cls):  # TODO 重写
        mode = cls.novelai_mode
        match mode:
            case "novelai":
                if cls.novelai_token:
                    cls.set_novelai()
                else:
                    logger.error(f"使用novelai模式但未检测到novelai token")
            case "naifu":
                if cls.novelai_api_domain == "https://api.novelai.net/" or cls.novelai_site_domain == "https://novelai.net/":
                    logger.error(f"请配置个人服务器api和site")
            case "webui":
                if cls.novelai_api_domain == "https://api.novelai.net/" or cls.novelai_site_domain == "https://novelai.net/":
                    logger.error(f"请配置个人服务器api和site")
            case _:
                if cls.novelai_token:
                    logger.error(f"请配置正确的运行模式，检测到token，自动切换至novelai模式")
                    cls.set_novelai()
                    cls.mode = "novelai"
                else:
                    logger.error(f"请配置正确的运行模式")

    def set_novelai(cls):
        cls.novelai_api_domain: str = "https://api.novelai.net/"
        cls.novelai_site_domain: str = "https://novelai.net/"

    async def set_enable(cls, group_id, enable):
        # 设置分群启用
        await cls.__init_json()
        now = await cls.get_value(group_id, "on")
        if now:
            if enable:
                return f"aidraw已经处于启动状态"
            else:
                await cls.set_value(group_id, "on","False")
                return f"aidraw已关闭"
        else:
            if enable:
                await cls.set_value(group_id, "on","True")
                return f"aidraw开始运行"
            else:
                return f"aidraw已经处于关闭状态"

    async def __init_json(cls):
        # 初始化设置文件
        if not jsonpath.exists():
            jsonpath.parent.mkdir(parents=True, exist_ok=True)
            async with aiofiles.open(jsonpath, "w+")as f:
                await f.write("{}")

    async def get_value(cls, group_id, arg: str):
        # 获取设置值
        group_id = str(group_id)
        arg_ = arg if arg.startswith("novelai_") else "novelai_"+arg
        if arg_ in cls.keys():
            await cls.__init_json()
            async with aiofiles.open(jsonpath, "r") as f:
                jsonraw = await f.read()
                configdict: dict = json.loads(jsonraw)
                return configdict.get(group_id, {}).get(arg_, False) or dict(cls)[arg_]
        else:
            return None

    async def get_groupconfig(cls, group_id):
        # 获取当群所有设置值
        group_id = str(group_id)
        await cls.__init_json()
        async with aiofiles.open(jsonpath, "r") as f:
            jsonraw = await f.read()
            configdict: dict = json.loads(jsonraw)
            baseconfig = {}
            for i in cls.keys():
                value = configdict.get(group_id, {}).get(
                    i, False) or dict(cls)[i]
                baseconfig[i] = value
            return baseconfig

    async def set_value(cls, group_id, arg: str, value: str):
        """设置当群设置值"""
        # 将值转化为bool和int
        if value.isdigit():
                value: int = int(value)
        elif value.lower() == "false":
                value = False
        elif value.lower() == "true":
                value = True
        group_id = str(group_id)
        arg_ = arg if arg.startswith("novelai_") else "novelai_"+arg
        # 判断是否合法
        if arg_ in cls.keys() and isinstance(value, type(dict(cls)[arg_])):
            await cls.__init_json()
            # 读取文件
            async with aiofiles.open(jsonpath, "r") as f:
                jsonraw = await f.read()
                configdict: dict = json.loads(jsonraw)
            # 设置值
            groupdict = configdict.get(group_id, {})
            if value == "default":
                groupdict[arg_] = False
            else:
                groupdict[arg_] = value
            configdict[group_id] = groupdict
            # 写入文件
            async with aiofiles.open(jsonpath, "w") as f:
                jsonnew = json.dumps(configdict)
                await f.write(jsonnew)
            return True
        else:
            logger.debug(f"不正确的赋值")
            return False


config = Config(**get_driver().config.dict())
logger.info(f"加载config完成" + str(config))



import aiohttp
import base64
from nonebot import on_command
from nonebot.adapters.onebot.v11 import GroupMessageEvent, MessageSegment
from nonebot.log import logger
from .translation import translate

deepdanbooru = on_command(".gettag", aliases={"鉴赏", "查书"})


@deepdanbooru.handle()
async def deepdanbooru_handle(event: GroupMessageEvent):
    url = ""
    for seg in event.message['image']:
        url = seg.data["url"]
    if url:
        async with aiohttp.ClientSession() as session:
            logger.info(f"正在获取图片")
            async with session.get(url) as resp:
                bytes = await resp.read()
        str_img = str(base64.b64encode(bytes), "utf-8")
        message = MessageSegment.at(event.user_id)
        start = "data:image/jpeg;base64,"
        str0 = start+str_img
        async with aiohttp.ClientSession() as session:
            async with session.post('https://mayhug-rainchan-anime-image-label.hf.space/api/predict/', json={"data": [str0, 0.6,"ResNet101"]}) as resp:
                if resp.status != 200:
                    await deepdanbooru.finish(f"识别失败，错误代码为{resp.status}")
                jsonresult = await resp.json()
                data = jsonresult['data'][0]
                logger.info(f"TAG查询完毕")
                tags = ""
                for label in data['confidences']:
                    tags = tags+label["label"]+","
        tags_ch = await translate(tags.replace("_", " "), "zh")
        if tags_ch == tags.replace("_", " "):
            message = message+tags
        message = message+tags+f"\n机翻结果:"+tags_ch
        await deepdanbooru.finish(message)
    else:
        await deepdanbooru.finish(f"未找到图片")



from dataclasses import dataclass, field
from io import BytesIO
from PIL import Image
import base64
from .config import config
from nonebot import get_driver
from .data import lowQuality
header = {
    "authorization": "Bearer " + config.novelai_token,
    ":authority": config.novelai_api_domain,
    ":path": "/ai/generate-image",
    "content-type": "application/json",
    "referer": config.novelai_site_domain,
    "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/106.0.0.0 Safari/537.36",
}

#自动切换模型
def set_model():
    if config.novelai_h:
        return "nai-diffusion"
    else:
        return "safe-diffusion"


@dataclass
class FIFO():
    """队列中的单个请求"""
    user_id: int
    tags: str
    seed: int
    group_id: int = 0
    scale: int = 11
    strength: float = 0.7
    samper: str = "k_euler_ancestral"
    count: int = 1
    model: str = field(default_factory=set_model)
    cost: int = 0
    steps: int = 28
    noise: float = 0.2
    uc: str = ""
    width: int = 512
    height: int = 512
    image: str = field(default=None, repr=False)

    def __post_init__(self):
        #数值合法检查
        if self.steps <= 0 or self.steps > 50:
            self.steps = 28
        self.uc = lowQuality+self.uc
        self.__image_check()
        #计算cost
        if config.novelai_paid == 1:
            anlas = 0
            if (self.width*self.height > 409600) or self.image or self.count > 1:
                anlas += round(self.width*self.height *
                               self.strength*self.count*self.steps/2293750)
            if self.user_id in get_driver().config.superusers:
                self.cost = 0
            else:
                self.cost = anlas
        else:
            self.cost = 0

    def __image_check(self):
        #根据图片重写长宽
        if self.image is not None:
            tmpfile = BytesIO(self.image)
            image = Image.open(tmpfile)
            width, height = image.size
            if width >= height:
                self.width = round(width/height*8)*64
                self.height = 512
            else:
                self.height = round(height/width*8)*64
                self.width = 512
            self.image = str(base64.b64encode(self.image), "utf-8")
            self.steps = 50

    def body(self):
        #获取请求体
        parameters = {
            "width": self.width,
            "height": self.height,
            "qualityToggle": False,
            "scale": self.scale,
            "sampler": self.samper,
            "steps": self.steps,
            "seed": self.seed,
            "n_samples": self.count,
            "ucPreset": 0,
            "uc": self.uc,
        }
        if self.image is not None:
            parameters.update({
                "image": self.image,
                "strength": self.strength,
                "noise": self.noise
            })
        return {
            "input": self.tags,
            "model": self.model,
            "parameters": parameters
        }
    def keys(self):
        return ("seed","tags","uc","scale","strength","noise","samper","model","steps","width","height")
    def __getitem__(self,item):
        return getattr(self,item)
    def format(self):
        dict_self=dict(self)
        str=""
        for i,v in dict_self.items():
            str+=f"{i}={v}\n"
        return str


from io import BytesIO
from PIL import Image
import re
import aiohttp
import os
from pathlib import Path


def is_contain_chinese(check_str: str):
    # 检查字符串是否包含汉字
    for ch in check_str:
        if u'\u4e00' <= ch <= u'\u9fff':
            return True
    return False


def file_name_check(input: str):
    # 检查文件名是否包含奇怪字符
    errorcode = ["?", "<", ">", "\\", "/", "*", "|", ":"]
    for i in errorcode:
        input.replace(i, "")
    return input


async def check_last_version(package: str):
    # 检查包的最新版本
    async with aiohttp.ClientSession() as session:
        async with session.get("https://pypi.org/simple/"+package) as resp:
            text = await resp.text()
            pattern = re.compile("-(\d.*?).tar.gz")
            pypiversion = re.findall(pattern, text)[-1]
    return pypiversion


async def compare_version(old: str, new: str):
    # 比较两个版本哪个最新
    oldlist = old.split(".")
    newlist = new.split(".")
    for i in range(len(oldlist)):
        if int(newlist[i]) > int(oldlist[i]):
            return True
    return False


async def sendtosuperuser(message):
    # 将消息发送给superuser
    from nonebot import get_bot, get_driver
    import asyncio
    superusers = get_driver().config.superusers
    bot = get_bot()
    for superuser in superusers:
        await bot.call_api('send_msg', **{
            'message': message,
            'user_id': superuser,
        })
        await asyncio.sleep(5)

async def png2jpg(raw: BytesIO):
    img_PIL = Image.open(raw)
    img_PIL.convert("RGB")
    image_new = BytesIO()
    img_PIL.save(image_new, format="JPEG", quality=95)
    return image_new



import aiohttp
import base64
from .utils import png2jpg
from io import BytesIO
from .fifo import FIFO,header
from .config import config
async def post(fifo:FIFO):
    #novelai请求交互
    img_bytes = []
    async with aiohttp.ClientSession(headers=header) as session:
        #向novelai服务器发送请求
        async with session.post(
            config.novelai_api_domain+"ai/generate-image", json=fifo.body()
        ) as resp:
            if resp.status != 201:
                return f"生成失败，错误代码为{resp.status}"
            img = await resp.text()
            img=img.split("data:")[1]

            #将图片转化为jpg(BytesIO)
            image=BytesIO(base64.b64decode(img))
            image_new=await png2jpg(image)
            img_bytes.append(image_new)
    return img_bytes



import base64
from io import BytesIO
import time
from pathlib import Path
import aiofiles
import aiohttp
import hashlib
import re
import asyncio
from nonebot import get_bot, get_driver, on_command
from nonebot.adapters.onebot.v11 import GroupMessageEvent, Message, MessageSegment, Bot
from nonebot.adapters.onebot.v11 import GROUP_ADMIN, GROUP_OWNER
from nonebot.log import logger
from nonebot.params import CommandArg
from .config import config
from .utils import is_contain_chinese, file_name_check
from .utils.translation import translate
from .version import version
from .utils.anlas import anlas_check, anlas_set
from .fifo import FIFO
from .data import htags
from .novelai import post
path = Path("data/novelai/output").resolve()
txt2img = on_command(".aidraw", aliases={"绘画", "咏唱", "约稿", "召唤"})

cd = {}
gennerating = False
limit_list = []
nickname = ""
for i in get_driver().config.nickname:
    nickname = i


@txt2img.handle()
async def txt2img_handle(bot: Bot, event: GroupMessageEvent, args: Message = CommandArg()):
    message_raw = args.extract_plain_text().replace("，", ",").split("-")
    user_id = str(event.user_id)
    count = 1
    # 以图生图预处理
    img_url = ""
    for seg in event.message['image']:
        img_url = seg.data["url"]
    imgbytes = ""
    if img_url:
        if config.novelai_paid:
            async with aiohttp.ClientSession() as session:
                logger.info(f"检测到图片，自动切换到以图生图，正在获取图片")
                async with session.get(img_url) as resp:
                    imgbytes = await resp.read()
        else:
            await txt2img.finish(f"以图生图功能已禁用")
    # 处理管理指令
    managetag = None
    managelist = message_raw[0].split()
    match managelist:
        case ["off"]:
            managetag = False
        case ["on"]:
            managetag = True
        case ["set"]:
            group_config = await config.get_groupconfig(event.group_id)
            message = "当前群的设置为\n"
            for i, v in group_config.items():
                message += f"{i}:{v}\n"
            await txt2img.finish(message)
        case ["set", arg, *value]:
            value_str=""
            for i in value:
                value_str=value_str+i+" "
            value=value_str
            if await GROUP_ADMIN(bot, event) or await GROUP_OWNER(bot, event) or event.user_id in get_driver().config.superusers:
                await txt2img.finish(f"设置群聊{arg}为{value}完成" if await config.set_value(event.group_id, arg, value) else f"不正确的赋值")
            else:
                await txt2img.finish(f"只有管理员可以使用管理功能")
    if managetag is not None:
        if await GROUP_ADMIN(bot, event) or await GROUP_OWNER(bot, event) or event.user_id in get_driver().config.superusers:
            result = await config.set_enable(event.group_id, managetag)
            logger.info(result)
            await txt2img.finish(result)
        else:
            await txt2img.finish(f"只有管理员可以使用管理功能")
    # 判断是否禁用，若没禁用，进入处理流程
    if await config.get_value(event.group_id, "on") is not None:
        # 判断cd
        nowtime = time.time()
        deltatime = nowtime - cd.get(user_id, 0)
        cd_ = int(await config.get_value(event.group_id, "cd"))
        if (deltatime) < cd_:
            await txt2img.finish(f"你冲的太快啦，请休息一下吧，剩余CD为{cd_-int(deltatime)}s")
        else:
            cd[user_id] = nowtime

        width, height = [512, 768]  # w*h
        tags = ""
        seed_raw = None
        nopre = False

        # 提取参数
        for i in message_raw:
            i = i.strip()
            match i:
                case "square" | "s":
                    width, height = [640, 640]
                case "portrait" | "p":
                    width, height = [512, 768]
                case "landscape" | "l":
                    width, height = [768, 512]
                case "nopre" | "np":
                    nopre = True
                case _:
                    if i.isdigit():
                        seed_raw = int(i)
                    else:
                        tags += i
        # 处理奇奇怪怪的输入
        tags = re.sub("[\f\n\r\t\v]", "", tags)
        tags = file_name_check(tags)
        # 生成种子
        seed = seed_raw or int(time.time())

        # 检测中文
        taglist = tags.split(",")
        tagzh = ""
        tags_ = ""
        for i in taglist:
            if is_contain_chinese(tags):
                tagzh += f"{i},"
            else:
                tags_ += f"{i},"
        if tagzh:
            tags_en = await translate(tagzh, "en")
            if tags_en == tagzh:
                await txt2img.finish(f"检测到中文，翻译失败，生成终止，请联系BOT主查看后台")
            else:
                tags_ += tags_en
        # 检测是否有18+词条
        if not config.novelai_h:
            pattern = re.compile(f"(\s|,|^)({htags})(\s|,|$)")
            if (re.search(pattern, tags) is not None):
                await txt2img.finish("H是不行的!")
        # 初始化队列
        tags = await config.get_value(event.group_id, "tag") + "," + tags_
        if imgbytes:
            fifo = FIFO(user_id, tags, seed, event.group_id, image=imgbytes)
        else:
            fifo = FIFO(user_id, tags, seed, event.group_id,
                        width=width, height=height)
        logger.debug(f"获取到请求{fifo}")
        if fifo.cost > 0:
            anlascost = fifo.cost
            hasanlas = await anlas_check(fifo.user_id)
            if hasanlas >= anlascost:
                await wait_fifo(fifo, anlascost, hasanlas-anlascost)
            else:
                await txt2img.finish(f"你的点数不足，你的剩余点数为{hasanlas}")
        else:
            await wait_fifo(fifo)


async def wait_fifo(fifo, anlascost=None, anlas=None):
    # 创建队列
    list_len = get_wait_num()
    has_wait = f"排队中，你的前面还有{list_len}人"
    no_wait = "请稍等，图片生成中"
    if anlas:
        has_wait += f"\n本次生成消耗点数{anlascost},你的剩余点数为{anlas}"
        no_wait += f"\n本次生成消耗点数{anlascost},你的剩余点数为{anlas}"
    if config.novelai_limit:
        await txt2img.send(has_wait if list_len > 0 else no_wait)
        limit_list.append(fifo)
        await fifo_gennerate()
    else:
        await txt2img.send(no_wait)
        await fifo_gennerate(fifo)


def get_wait_num():
    # 获取剩余队列长度
    list_len = len(limit_list)
    if gennerating:
        list_len += 1
    return list_len


async def fifo_gennerate(fifo: FIFO = None):
    # 队列处理
    global gennerating
    bot = get_bot()

    async def generate(fifo: FIFO):
        # 开始生成
        logger.info(
            f"队列剩余{get_wait_num()}人 | 开始生成：{fifo}")
        try:
            im = await _run_gennerate(fifo)
        except:
            logger.exception("生成失败")
            await bot.send_group_msg(
                message="生成失败，请联系BOT主排查原因",
                group_id=fifo.group_id
                )
        else:
            logger.info(f"队列剩余{get_wait_num()}人 | 生成完毕：{fifo}")

        if await config.get_value(fifo.group_id,"pure"):
            message=MessageSegment.at(fifo.user_id)
            for i in im["image"]:
                message+=i
            await bot.send_group_msg(
                message=message,
                group_id=fifo.group_id,
            )
        else:
            message=[]
            for i in im:
                message.append(MessageSegment.node_custom(bot.self_id,nickname,i))
            await bot.send_group_forward_msg(
                messages=message,
                group_id=fifo.group_id,
            )

    if fifo:
        await generate(fifo)

    if not gennerating:
        logger.info("队列开始")
        gennerating = True

        while len(limit_list) > 0:
            fifo = limit_list.pop(0)
            try:
                await generate(fifo)
            except:
                logger.exception("生成中断")

        gennerating = False
        logger.info("队列结束")
        await version.check_update()


async def _run_gennerate(fifo: FIFO):
    # 处理单个请求
    img_bytes = await post(fifo)
    message = MessageSegment.text(fifo.format())
    # 判断是否允许H
    if config.novelai_h:
        for i in img_bytes:
            await save_img(fifo, i)
            message += MessageSegment.image(i)
    else:
        nsfw_count = 0
        for i in img_bytes:
            try:
                label = await check_safe(i)
            except RuntimeError as e:
                logger.error(f"NSFWAPI调用失败，错误代码为{e.args}")
                label = "unknown"
            if label != "explicit":
                message += MessageSegment.image(i)
            else:
                nsfw_count += 1
            await save_img(fifo, i, label)
        if nsfw_count>0:
            message += f"\n有{nsfw_count}张图片太涩了，{nickname}已经帮你吃掉了哦"
    #扣除点数
    if fifo.cost > 0:
        await anlas_set(fifo.user_id, -fifo.cost)
    return message


async def save_img(fifo, img_bytes:BytesIO, extra: str = "unknown"):
    #存储图片
    if config.novelai_save_pic:
        path_ = path/extra
        path_.mkdir(parents=True, exist_ok=True)
        hash=hashlib.md5(img_bytes.getvalue()).hexdigest()
        file = (path_/hash).resolve()
        async with aiofiles.open(str(file)+".jpg", "wb") as f:
            await f.write(img_bytes.getvalue())
        async with aiofiles.open(str(file)+".txt","w") as f:
            await f.write(fifo.format())


async def check_safe(img_bytes:BytesIO):
    #检查图片是否安全
    start = "data:image/jpeg;base64,"
    image=img_bytes.getvalue()
    image=str(base64.b64encode(image), "utf-8")
    str0 = start+image
    #重试三次
    for i in range(3):
        async with aiohttp.ClientSession() as session:
            #调用API
            async with session.post('https://hf.space/embed/mayhug/rainchan-image-porn-detection/api/predict/', json={"data": [str0]}) as resp:
                if resp.status == 200:
                    jsonresult = await resp.json()
                    break
                else:
                    await asyncio.sleep(2)
                    error = resp.status
    else:
        raise RuntimeError(error)
    return jsonresult["data"][0]["label"]



# 基础优化tag
basetag = "masterpiece, best quality,"

# 基础排除tag
lowQuality = "lowres, bad anatomy, bad hands, text, error, missing fingers, extra digit, fewer digits, cropped, worst quality, low quality, normal quality, jpeg artifacts, signature, watermark, username, blurry, pubic hair,long neck,blurry"

# 屏蔽词
htags = "nsfw|nude|naked|nipple|blood|censored|vagina|gag|gokkun|hairjob|tentacle|oral|fellatio|areolae|lactation|paizuri|piercing|sex|footjob|masturbation|hips|penis|testicles|ejaculation|cum|tamakeri|pussy|pubic|clitoris|mons|cameltoe|grinding|crotch|cervix|cunnilingus|insertion|penetration|fisting|fingering|peeing|ass|buttjob|spanked|anus|anal|anilingus|enema|x-ray|wakamezake|humiliation|tally|futa|incest|twincest|pegging|femdom|ganguro|bestiality|gangbang|3P|tribadism|molestation|voyeurism|exhibitionism|rape|spitroast|cock|69|doggystyle|missionary|virgin|shibari|bondage|bdsm|rope|pillory|stocks|bound|hogtie|frogtie|suspension|anal|dildo|vibrator|hitachi|nyotaimori|vore|amputee|transformation|bloody"
